# -*- coding:utf-8 -*-
import datetime

import const

# 当mysql源库参数binlog_row_image设置为minimal时，
# binlog中update记录的未修改字段的值是缺失的，
# 这时用这个特定字符串代表缺失字段的值。
JSON_NA_VALUE = "##DTS_NA_VALUE##"


def trans2sql(record):
    message_type = record[const.MESSAGE_TYPE]

    if message_type == const.HEARTBEAT:
        print("heartbeat:{}".format(record[const.HEARTBEAT_TIMESTAMP]))
    elif message_type == const.CHECKPOINT:
        print("Checkpoint: gtid={}, file={}, position={}".format(record[const.SYNCED_GTID],
                                                                 record[const.FILE_NAME],
                                                                 record[const.POSITION]))
    elif message_type == const.BEGIN:
        print("BEGIN")
    elif message_type == const.COMMIT:
        print("COMMIT")
    elif message_type == const.ROLLBACK:
        print("ROLLBACK")
    elif message_type == const.DDL:
        print(record[const.SQL])
    elif message_type == const.INSERT:
        print(trans_insert(record))
    elif message_type == const.DELETE:
        print(trans_delete(record))
    elif message_type == const.UPDATE:
        print(trans_update(record))
    elif message_type ==  const.INIT_INSERT:
        print(trans_init_insert(record))
    elif message_type == const.INIT_DDL:
        print(record[const.SQL]) 
    elif message_type == const.STATEMENT:
        print(record[const.SQL])
    else:
        print("unsupported message type")


def trans_insert(record) -> str:
    col_values = []

    for idx, col in enumerate(record[const.NEW_COLUMNS]):
        d = decode(col, record[const.COLUMNS][idx])
        col_values.append(d)

    sql = "INSERT INTO `{}`.`{}` VALUES ({})".format(
        record[const.SCHEMA_NAME],
        record[const.TABLE_NAME],
        ", ".join(col_values))

    return sql

def trans_init_insert(record) -> str:
    col_values = []
    col_names = []

    for col in record[const.COLUMNS]:
        col_names.append(col['name'])

    for idx, col in enumerate(record[const.NEW_COLUMNS]):
        d = decode(col, record[const.COLUMNS][idx])
        col_values.append(d)

    sql = "INSERT INTO `{}`.`{}`({}) VALUES ({})".format(
        record[const.SCHEMA_NAME],
        record[const.TABLE_NAME],
        ", ".join(col_names),
        ", ".join(col_values))

    return sql

def trans_delete(record) -> str:
    conditions = []

    for idx, col in enumerate(record[const.OLD_COLUMNS]):
        col_def = record[const.COLUMNS][idx]
        d = decode(col, col_def)

        if "DEFAULT" == d or "NULL" == d:
            continue

        conditions.append("{} = {}".format(col_def['name'], d))

    where = " AND ".join(conditions)
    sql = "DELETE FROM `{}`.`{}` WHERE {} LIMIT 1".format(
        record[const.SCHEMA_NAME], record[const.TABLE_NAME], where)

    return sql


def trans_update(record) -> str:
    assignments = []

    for idx, col in enumerate(record[const.NEW_COLUMNS]):
        if col == JSON_NA_VALUE:
            continue

        col_def = record[const.COLUMNS][idx]
        d = decode(col, col_def)
        assignments.append("{} = {}".format(col_def['name'], d))

    conditions = []
    for idx, col in enumerate(record[const.OLD_COLUMNS]):
        col_def = record[const.COLUMNS][idx]
        d = decode(col, col_def)
        if "DEFAULT" == d or "NULL" == d:
            continue

        conditions.append("{} = {}".format(col_def['name'], d))

    assign = ", ".join(assignments)
    where = " AND ".join(conditions)
    sql = "UPDATE `{}`.`{}` SET {} WHERE {} LIMIT 1".format(
        record[const.SCHEMA_NAME], record[const.TABLE_NAME], assign, where)

    return sql


def decode(data, field) -> str:
    """data有以下几种类型：
        none表示数据库的null值
        特殊字符串##DTS_NA_VALUE##表示缺失字段的值（binlog_row_image为minimal）
        其他正常值为字符串，分为两种：
        （1）16进制字符串表示mysql中的二进制类型，如blob，binary，geometry等；
        （2）mysql中的其他数据类型用字符串记录真实值。
    """
    if data is None:
        return "NULL"

    if isinstance(data, str):
        if JSON_NA_VALUE == data:
            return "DEFAULT"

        if "binary" in field['originalType'].lower() or \
                "blob" in field['originalType'].lower() or \
                const.MYSQL_TYPE_GEOMETRY == field['dataTypeNumber']:
            return "x'" + data + "'"

        if field['dataTypeNumber'] in (const.MYSQL_TYPE_DECIMAL, const.MYSQL_TYPE_DECIMAL_NEW,
                                       const.MYSQL_TYPE_INT8, const.MYSQL_TYPE_INT16, const.MYSQL_TYPE_INT24,
                                       const.MYSQL_TYPE_INT32, const.MYSQL_TYPE_INT64, const.MYSQL_TYPE_FLOAT,
                                       const.MYSQL_TYPE_DOUBLE, const.MYSQL_TYPE_BIT, const.MYSQL_TYPE_NULL):
            return str(data)

        if field['dataTypeNumber'] == const.MYSQL_TYPE_JSON:
            return "'" + data + "'"

        if field['dataTypeNumber'] in (const.MYSQL_TYPE_TIMESTAMP, const.MYSQL_TYPE_TIMESTAMP_NEW):
            # 时间戳返回的毫秒数
            timestamp = int(data)
            dt = datetime.datetime.fromtimestamp(timestamp=timestamp / 1000)
            return "_binary'" + dt.strftime('%Y-%m-%d %H:%M:%S.%f') + "'"

        if field['dataTypeNumber'] in (const.MYSQL_TYPE_DATETIME, const.MYSQL_TYPE_DATETIME_NEW):
            tstr = data.split(" ")
            if len(tstr) == 3:
                # datetime 可能带时区信息
                return "_binary'" + tstr[0] + " " + tstr[1] + "'"
            return "_binary'" + data + "'"

        return "_binary'" + data + "'"
    else:
        print("unknown value: {}".format(data))


if __name__ == '__main__':
    s: str = "ding"
    print("ing" in s.lower())
